#define cimg_display 0
#include "include/CImg.h"
#include "include/Vector.h"

#include <fstream>
#include <vector>
#include <string>
using namespace cimg_library;
using namespace std;
using namespace base;

#define SQR(x) ((x)*(x))
#define MAX(x,y) ((x)>(y)?(x):(y))
#define MAX_BUMPS 10000
#define NUM_ATTEMPTS 100000
#define BP 7

#define PAD (2*radius)
#define OS 1

// All these variables are scriptable if you run the executable with a mdef file
Vec3f light1_color(1,0.8,0.7);
Vec3f light1_dir(-0.3,1,0.2);

Vec3f light2_color(0.4,0.0,0.0);
Vec3f light2_dir(0,-0.5,-0.05);

Vec3f light3_ambientcolor(0.05,0.05,0.05);

Vec3i material1(0xCC, 0x99, 0x33);
Vec3f diffuse_albedo_material1(0.4, 0.2, 0.1);
float specular_exponent_material1 = 30;
float specular_intens_material1 = 0.05;
Vec3f specular_color_material1(1,1,1);

Vec3i material2(0xb7, 0xb7, 0xb7);
Vec3f diffuse_albedo_material2(0.1, 0.1, 0.13);
float specular_exponent_material2 = 20;
float specular_intens_material2 = 0.7;
Vec3f specular_color_material2(1,1,1);

Vec3f diffuse_albedo1_cover1(0.2,0.8,0.0);
Vec3f diffuse_albedo2_cover1(0.1,0.4,0.0);
float specular_exponent_cover1 = 10;
float specular_intens_cover1 = 0.0;

Vec3f diffuse_albedo1_cover2(0.1,0.4,0.0);
Vec3f diffuse_albedo2_cover2(0.05,0.2,0.0);
float specular_exponent_cover2 = 10;
float specular_intens_cover2 = 0.0;

string environment="";
float environment_gain = 1.f;
char environment_flip = ' ';

float bumps[BP] = {5,10,1,12,1,0,0};
bool grass = false;

int radius = 7;
int sigma = 1;

// Bump definition:
// [0]: float min_size
// [1]: float max_size
// [2]: float density (overfill factor - values > 1 will lead to overlap)
// [3]: float amplitude (how much height do the bumps add?)
// [4]: int algo (sampling algorithm. 0: closer to blue noise; 1: clusters of small ones
// [5]: int shape (shape of bumps. 0: Gaussian, 1: craters, 2: sharp bumps
// [6]: int blendmode (how do bumps combine? 0: additive, 1: MAX operator) 

typedef struct bumpdef {float x[BP];} bumpdef;

vector<bumpdef> bumplayers;


float cover_thres = 0.2; // How "horizontal" does a surface need to be for grass to grow on it? (1 = absolutely flat; 0 = anything that's not absolutely vertical

int pattern_cover = 1; // 0=grass;1=stripes;2=checkers
int thickness_cover = 5; // thickness of cover layer in pixels


// Extremely inefficient generator of blue noise  (rejection sampling)
int bluenoise(int dimx, int dimy, float min_dist, float max_dist, float* x, float* y, float* r, int algo=1, float density = 1) {

	
	int i = 0;
	float nr = max_dist;
	float acceptance_rate = 1;
	
	for (int attempts = 0; i < MAX_BUMPS && attempts < NUM_ATTEMPTS && nr > min_dist; ++attempts) {
		// draw new position randomly
		float nx = (float) rand() / RAND_MAX * dimx;
		float ny = (float) rand() / RAND_MAX * dimy;
		
		switch(algo) {
		
		case 0: 
		// Option 0: Start with blue-noise sampling of large dots, then decrease size as acceptance rate goes down
				if (acceptance_rate < 0.001) { nr *= 0.99; acceptance_rate = 1;}
				break;				
		case 1: 
		// Option 1: Always randomly pick size of dots. Will lead to clusters of small bumps
		default: nr = (float) rand() / RAND_MAX * (max_dist-min_dist) + min_dist;
		}

		// and if it's too close to samples drawn so far, reject
		bool valid = true;
		for (int j = 0; valid && j < i; ++j) 
			if (SQR(x[j]-nx)+SQR(y[j]-ny)<SQR(r[j]+nr)/density) valid = false;
	
		if (valid) { 
			x[i] = nx; y[i] = ny; r[i] = nr; ++i;
		}
		acceptance_rate = 0.95*acceptance_rate + 0.05*(valid?1:0);
	}
	return i;
}

// Float-valued function that gives "bump" profiles of size r and a choice of different shapes
float profile(float xx, float yy, float r, int type) {

float r_rel2 = (SQR((float)xx)+SQR((float)yy))/SQR(r);
switch(type) {
case 1: 
return exp(-abs(sqrt(r_rel2)-0.7)*4);				// Gaussian craters
case 2: 
return 1-exp(-MAX(0.f,0.7-sqrt(r_rel2))*4);			// Sharp-edged bumps
default: 
case 0: return exp(-r_rel2);						// Gaussians
}
}

// Generate bumps in a height map
void bump(CImg<float> &im, bumpdef &bumps) {
	
	float min_size = bumps.x[0]; 
	float max_size = bumps.x[1];
	float density = bumps.x[2];
	float amplitude = bumps.x[3];
	int algo = bumps.x[4];
	int type = bumps.x[5];
	int compose = bumps.x[6];

	float x[MAX_BUMPS],y[MAX_BUMPS],r[MAX_BUMPS];
	int num_bumps = bluenoise(im.width(),im.height(),min_size*OS,max_size*OS,x,y,r,algo,density);
	cout << "("<<num_bumps<<" bumps)";
	
	switch(compose) {
	case 0: // additive composition
	for (int i = 0; i < num_bumps; ++i) {
		int radius = 3*r[i];
		for (int yy = -radius; yy < radius; ++yy) 
		for (int xx = -radius; xx < radius; ++xx) {
		im.atXY(x[i]+xx,y[i]+yy,0,0) += amplitude*r[i]*profile(xx,yy,r[i],type);
		}	
		//im.atXY(x[i],y[i],0,2)=255;
	}; break;
	case 1: // MAX operator
	for (int i = 0; i < num_bumps; ++i) {
		int radius = 3*r[i];
		for (int yy = -radius; yy < radius; ++yy) 
		for (int xx = -radius; xx < radius; ++xx) {
		im.atXY(x[i]+xx,y[i]+yy,0,0) = MAX(im.atXY(x[i]+xx,y[i]+yy,0,0),amplitude*r[i]*profile(xx,yy,r[i],type));
		}	
		//im.atXY(x[i],y[i],0,2)=255;
	}; break;
	
	default: ;// Do nothing
	}
}

CImg<unsigned char> render(CImg<float> normals, CImg<unsigned char> level, CImg<float> env, float mul=1, float gamma=2.2) {

//  Normalize vectors
	light1_dir.normalize();
	light2_dir.normalize();

	CImg<float> im(normals);
	im.fill(0);
	float max_n = 0;
	
	// Find maximum normal vector component, and normalize
	for (int x = PAD*OS; x < normals.width()-PAD*OS; ++x)
		for (int y =  PAD*OS; y < normals.height()- PAD*OS; ++y)
			max_n = MAX(MAX(max_n,abs(normals(x,y,0,0))),abs(normals(x,y,0,1))); 
			
	cout << "max_n = "<<max_n<<endl;		

	normals /= max_n; // range should now go from -1.f .. 1.f

	for (int x = 0; x < normals.width(); ++x)
		for (int y = 0; y < normals.height(); ++y) {
				float remainder = 1-SQR(normals(x,y,0,0))-SQR(normals(x,y,0,1));
				if (remainder >= 0) 
					normals(x,y,0,2) = sqrt(remainder);
				else 
					normals(x,y,0,2) = 0;
				
				// and post-normalize:
				float norm = sqrt(SQR(normals(x,y,0,0))+SQR(normals(x,y,0,1))+SQR(normals(x,y,0,2)));
				normals(x,y,0,0) /= norm;
				normals(x,y,0,1) /= norm;
				normals(x,y,0,2) /= norm;
				}
	
cout << "Materials...";	

	if (environment == "") {
	// Render pixels of material 1 and 2
	for (int x = 0; x < normals.width(); ++x)
		for (int y = 0; y < normals.height(); ++y) {
		Vec3uc color(level(x,y,0,0),level(x,y,0,1),level(x,y,0,2));
		Vec3f n(normals(x,y,0,0),normals(x,y,0,1),normals(x,y,0,2));
		Vec3f diffuse(0,0,0);
		Vec3f specular(0,0,0);
		Vec3f Rm1 = 2*(light1_dir|n)*n-light1_dir;
		Vec3f Rm2 = 2*(light2_dir|n)*n-light2_dir;
		if (color==material1) { 
			 diffuse = diffuse_albedo_material1*(MAX(0,n|light1_dir)*light1_color + MAX(0,n|light2_dir)*light2_color+light3_ambientcolor);
			specular = light1_color * pow(MAX(0,Rm1[2]),specular_exponent_material1)
			+light2_color * pow(Rm2[2],specular_exponent_material2);
			specular *= specular_intens_material1;
		} else if (color[1] && color[1]==color[2] && color[1] == color[0]) { 
			 diffuse = diffuse_albedo_material2*(MAX(0,n|light1_dir)*light1_color + MAX(0,n|light2_dir)*light2_color+light3_ambientcolor);
			specular = light1_color * pow(MAX(0,Rm1[2]),specular_exponent_material2)
			+light2_color * pow(Rm2[2],specular_exponent_material2);
			specular *= specular_intens_material2;
		}
		im(x,y,0,0)=diffuse[0]+specular[0];
		im(x,y,0,1)=diffuse[1]+specular[1];
		im(x,y,0,2)=diffuse[2]+specular[2];

	}
	} else { // Environment lighting is a tad different
	int w = env.width();
	int h = env.height();
	for (int x = 0; x < normals.width(); ++x)
		for (int y = 0; y < normals.height(); ++y) {
		Vec3uc color(level(x,y,0,0),level(x,y,0,1),level(x,y,0,2));
		Vec3f n(normals(x,y,0,0),normals(x,y,0,1),normals(x,y,0,2));
		Vec3f diffuse(0,0,0);
		Vec3f specular(0,0,0);
		Vec3f spec_dir = Vec3f(w/2+n[0]*w/3,h/2+n[1]*h/3,0);
		if (color==material1) { 
			 diffuse = light3_ambientcolor * diffuse_albedo_material1;//diffuse_albedo_material1*(MAX(0,n|light1_dir)*light1_color + MAX(0,n|light2_dir)*light2_color+light3_ambientcolor);
			specular = environment_gain * Vec3f(env(spec_dir[0],spec_dir[1],0,0),
						env(spec_dir[0],spec_dir[1],0,1),
						env(spec_dir[0],spec_dir[1],0,2)); // */
			specular *= specular_color_material1*specular_intens_material1;
		} else if (color[1] && color[1]==color[2] && color[1] == color[0]) { 
			diffuse = light3_ambientcolor * diffuse_albedo_material1;//diffuse_albedo_material2*(MAX(0,n|light1_dir)*light1_color + MAX(0,n|light2_dir)*light2_color+light3_ambientcolor);
			specular = environment_gain * Vec3f(env(spec_dir[0],spec_dir[1],0,0),
						env(spec_dir[0],spec_dir[1],0,1),
						env(spec_dir[0],spec_dir[1],0,2));//*/
			specular *= specular_color_material2*specular_intens_material2;
		}
		im(x,y,0,0)=diffuse[0]+specular[0];
		im(x,y,0,1)=diffuse[1]+specular[1];
		im(x,y,0,2)=diffuse[2]+specular[2];
	}
		
	}
	cout << "done."<<endl;
	// Render cover pixels of material 1 and 2
	float n;
	
	if (grass) {
	cout << "Grass...";	
	for (int x = 0; x < normals.width(); ++x) {
		//float cover_thres = (float)rand() / RAND_MAX * 0.6+0.2;
		float blend;
		for (int y = 0; y < normals.height(); ++y) {
		if (y % 4 == rand() % 4) blend =  (float)rand() / RAND_MAX;
		Vec3uc color(level(x,y,0,0),level(x,y,0,1),level(x,y,0,2));
		Vec3f n(normals(x,y,0,0),normals(x,y,0,1),normals(x,y,0,2));

		if (n[1] > cover_thres && color == material1) {
		Vec3f diffuse;
		float alpha = (n[1]-cover_thres)/(1-cover_thres);
		Vec3f albedo;

			Vec3f	albedo1 = diffuse_albedo1_cover1 * alpha + diffuse_albedo2_cover1 * (1-alpha);
			Vec3f	albedo2 = diffuse_albedo1_cover2 * alpha + diffuse_albedo2_cover2 * (1-alpha);
albedo = albedo1*blend + albedo2*(1-blend);
			diffuse = albedo*(MAX(0,n|light1_dir)*light1_color + MAX(0,n|light2_dir)*light2_color+light3_ambientcolor);
		im(x,y,0,0)=diffuse[0];
		im(x,y,0,1)=diffuse[1];
		im(x,y,0,2)=diffuse[2];
			
	}
	}
	}
	cout << "done."<<endl;
	}
	im.save("/cygdrive/c/tmp/render.pfm");
	im=im.pow(1/gamma);
	im*=mul;
	im = im.min(255);
	CImg<unsigned char> io(im);
	return io;
	
}

void cmdline(int argc, char** argv){ 
	char fname[256];
	sprintf(fname,"style/%s.mdef",argv[3]);
	ifstream cin(fname);
	while (!cin.eof()) {
		std::string token;
		cin >> token;
		cout << token<< endl;
		if (token=="//") cin.ignore(1e6, '\n');
		else if (token=="light1_color") 
			cin >> light1_color[0] >> light1_color[1] >> light1_color[2];
		else if (token=="light1_dir") 
			cin >> light1_dir[0] >> light1_dir[1] >> light1_dir[2];
		else if (token=="light2_color") 
			cin >> light2_color[0] >> light2_color[1] >> light2_color[2];
		else if (token=="light2_dir") 
			cin >> light2_dir[0] >> light2_dir[1] >> light2_dir[2];
		else if (token=="light3_ambientcolor") 
			cin >> light3_ambientcolor[0] >> light3_ambientcolor[1] >> light3_ambientcolor[2];
		else if (token=="environment") {
			cin >> environment;
			cout << environment;
		} else if (token=="environment_gain") {
			cin >> environment_gain;
		} else if (token=="environment_flip") {
			cin >> environment_flip;
		} else if (token=="radius") 
			cin >> radius;
		else if (token=="material1") 
			cin >> material1[0] >> material1[1] >> material1[2];
		else if (token=="diffuse_albedo_material1") 
			cin >> diffuse_albedo_material1[0] >> diffuse_albedo_material1[1] >> diffuse_albedo_material1[2];
		else if (token=="specular_exponent_material1") 
			cin >> specular_exponent_material1;
		else if (token=="specular_intens_material1") 
			cin >> specular_intens_material1;
		else if (token=="specular_color_material1") 
			cin >> specular_color_material1[0] >> specular_color_material1[1] >> specular_color_material1[2];
		else if (token=="diffuse_albedo_material2") 
			cin >> diffuse_albedo_material2[0] >> diffuse_albedo_material2[1] >> diffuse_albedo_material2[2];
		else if (token=="specular_exponent_material2") 
			cin >> specular_exponent_material2;
		else if (token=="specular_intens_material2") 
			cin >> specular_intens_material2;
		else if (token=="specular_color_material2") 
			cin >> specular_color_material2[0] >> specular_color_material2[1] >> specular_color_material2[2];
		else if (token=="diffuse_albedo1_cover1") 
			cin >> diffuse_albedo1_cover1[0] >> diffuse_albedo1_cover1[1] >> diffuse_albedo1_cover1[2];
		else if (token=="diffuse_albedo2_cover1") 
			cin >> diffuse_albedo2_cover1[0] >> diffuse_albedo2_cover1[1] >> diffuse_albedo2_cover1[2];
		else if (token=="diffuse_albedo1_cover2") 
			cin >> diffuse_albedo1_cover2[0] >> diffuse_albedo1_cover2[1] >> diffuse_albedo1_cover2[2];
		else if (token=="diffuse_albedo2_cover2") 
			cin >> diffuse_albedo2_cover2[0] >> diffuse_albedo2_cover2[1] >> diffuse_albedo2_cover2[2];	
		else if (token=="bumps") {
			bumpdef bd;
			for (int i = 0; i < BP; ++i)
				cin >> bd.x[i];  
			bumplayers.push_back(bd);	
	} else if (token=="grass") grass = true;
}
};


int main(int argc, char** argv){

cmdline(argc,argv);

CImg<unsigned char> im(argv[1]);
int w = im.width();
int h = im.height();

CImg<unsigned char> io(im);
CImg<float> in;
in.assign<unsigned char>(im);

in.fill(0);
in.resize(w+2*PAD, h+2*PAD, 1,3,0,1,0.5,0.5);
im.resize(w+2*PAD, h+2*PAD, 1,3,0,1,0.5,0.5);


// Loop through all pixels, compute "land coverage" by Gaussian blur, add bump maps and compute normals

cout << "=== PASS 1: Filtering ===" << endl;
cout << "Segmenting...";
CImg<float> imb(im.width(),im.height(),1,1); // Binary image
CImg<float> imbb(im.width(),im.height(),1,1);// Greyscale (distance) image
// Binary img first
for (int x = 0; x < im.width(); ++x) {
	if (x%16==0) cout << "."<<flush;
	for (int y = 0; y < im.height(); ++y) {
			if (im(x,y,0,0) == material1[0] && im(x,y,0,1) == material1[1] && im(x,y,0,2) == material1[2] || im(x,y,0,0) && im(x,y,0,0)==im(x,y,0,1) && im(x,y,0,0)==im(x,y,0,2)) // it's a land pixel
			imb(x,y,0,0) = 1;
			else imb(x,y,0,0) = 0;
}}
cout << "done." << endl;
imb.save("/cygdrive/c/tmp/binary.pgm");

float norm = 0;
for (int xx = -sigma*radius; xx <= sigma*radius; ++xx){
		
			norm+=exp(-((float)xx*xx)/SQR(radius));
		
}
cout << "Gaussian horizontal...";
for (int x = 0; x < in.width(); ++x) {
	if (x%16==0) cout << "."<<flush;
	for (int y = 0; y < in.height(); ++y) {
		float sum = 0;
		for (int xx = -sigma*radius; xx <= sigma*radius; ++xx)
			sum+=imb.atXY(x+xx,y,0,0) * exp(-((float)xx*xx)/SQR(radius));
		imbb(x,y,0,0) = sum/norm;
	}
}
cout << "done." << endl;
cout << "Gaussian vertical...";
for (int x = 0; x < in.width(); ++x) {
	if (x%16==0) cout << "."<<flush;
	for (int y = 0; y < in.height(); ++y) {
		float sum = 0;
		for (int xx = -sigma*radius; xx <= sigma*radius; ++xx)
			sum+=imbb.atXY(x,y+xx,0,0) * exp(-((float)xx*xx)/SQR(radius));
		imb(x,y,0,0) = 384*pow(MAX(0,2*(sum/norm)),0.5);
	}
}
cout << "done." << endl;

in.resize(OS*(w+2*PAD),OS*(h+2*PAD),1,3,3);
im.resize(OS*(w+2*PAD),OS*(h+2*PAD),1,3,3);
imb.resize(OS*(w+2*PAD),OS*(h+2*PAD),1,1,3);

cout << "Adding bumps...";
for (int i = 0; i < bumplayers.size(); ++i) {
	cout << " " << i << ((i<bumplayers.size()-1)?',':' ') << flush;
	bump(imb,bumplayers[i]);
}
cout << "done." << endl;

cout << "Computing normals...";
for (int x = 0; x < in.width(); ++x) {
	if (x%16==0) cout << "."<<flush;
	for (int y = 0; y < in.height(); ++y) {
	in(x,y,0,0)=(imb.atXY(x-1,y,0,0)-imb.atXY(x+1,y,0,0))/2;
	in(x,y,0,1)=(imb.atXY(x,y+1,0,0)-imb.atXY(x,y-1,0,0))/2;
	}
}
cout << "done." << endl;
in.save("/cygdrive/c/tmp/normals.pfm");

CImg<float> env;
if (environment != "") { env.load((string("style/")+environment).c_str());
if (environment_flip != ' ')
	env.mirror(environment_flip);
}

// Unpad image
in.crop(OS*PAD,OS*PAD, OS*(PAD+w)-1, OS*(PAD+h)-1);
im.crop(OS*PAD,OS*PAD, OS*(PAD+w)-1, OS*(PAD+h)-1);

cout << "=== PASS 2: Rendering ===" << endl;
CImg<unsigned char> output = render(in,im,env,255,2.2);

#if (OS>1)
output.resize(w,h,1,3,3);
#endif
output.save(argv[2]);

return 0;
}
